from __future__ import unicode_literals

from rest_framework.pagination import PageNumberPagination
from .models import UniqueVote, ElectionVote
from django.db import transaction
from django.conf import settings
from django_redis import get_redis_connection

import string
import random


def select_random_entry(Model, election_id):
    """ Select a random empty slot to insert into a Model DB """

    optimize = getattr(settings, 'EVOTE_OPTIMIZED_INSERT', False)

    if(optimize):
        redis_connection = get_redis_connection("default")
        # Generate the necessary cache keys
        slots_cache_key = Model.__name__ + str(election_id) + "_slots"
        # Stored Min and Max values
        range_cache_key = Model.__name__ + str(election_id) + "_range"

        # Check if ranges are stored
        range_ids = redis_connection.smembers(range_cache_key)
        if range_ids:
            id_interval = range_ids.pop()
            (min_id, max_id) = id_interval.split(',')
            min_id = int(min_id)
            max_id = int(max_id)

            total_entries = (max_id - min_id)

            random_int = random.randrange(min_id, max_id)

            while(not redis_connection.sadd(slots_cache_key, random_int)):
                random_int = random.randrange(min_id, max_id)

            return (random_int, total_entries)

    min_entry = Model.objects.filter(
        election_id=election_id).order_by('id')[:1]

    max_entry = Model.objects.filter(
        election_id=election_id).order_by('-id')[:1]

    min_id = min_entry[0].id
    max_id = max_entry[0].id

    total_entries = (max_id - min_id)

    random_int = random.randrange(min_id, max_id)

    if(optimize):
        range_id = str(min_id) + ',' + str(max_id)
        redis_connection.sadd(range_cache_key, range_id)
        redis_connection.sadd(slots_cache_key, random_int)

    return (random_int, total_entries)


@transaction.atomic
def save_atomic_entry(obj, slot_id, election_id, max_nr):

    slot_entry = obj.__class__.objects.select_for_update().get(pk=slot_id)
    available_slots = obj.__class__.objects.filter(
        election_id=election_id, empty_entry=False).count()

    if(available_slots * settings.EVOTE_SHUFFLE_MULTIPLIER > max_nr):
        raise ValueError("Maximum number of votes reached")

    if not slot_entry.empty_entry:
        return False

    obj.id = slot_id
    obj.save()

    return slot_id


def initialize_election(election_id, cardinality):
    """ Initialize the election entries for the
        given election_id """
    # TODO: Mock lesser batch limit
    batch_limit = 3000
    multiplier = settings.EVOTE_SHUFFLE_MULTIPLIER
    votes = []
    bulletins = []
    for i in range(cardinality * multiplier):
        ev = ElectionVote(election_id=election_id,
                          empty_entry=True,
                          cyphered_vote='',
                          blind_components='',
                          pr_components='',
                          signed_vote_hash='')
        votes.append(ev)

        uv = UniqueVote(election_id=election_id,
                        empty_entry=True,
                        unique_vote_bulletin='')
        bulletins.append(uv)
        if(len(votes) > batch_limit):
            ElectionVote.objects.bulk_create(votes)
            UniqueVote.objects.bulk_create(bulletins)
            votes = []
            bulletins = []

    ElectionVote.objects.bulk_create(votes)
    UniqueVote.objects.bulk_create(bulletins)

    return True


def close_election(election_id):
    """ Delete dummy election entries for
        the given election_id """
    ElectionVote.objects.filter(empty_entry=True,
                                election_id=election_id).delete()
    UniqueVote.objects.filter(empty_entry=True,
                              election_id=election_id).delete()

    return True


def genstr(size=64,
           prefix='autogen_',
           chars=string.ascii_uppercase + string.digits):
    rnd_str = prefix
    for i in range(size - len(prefix)):
        rnd_str += random.choice(chars)

    return rnd_str


class StandardPagination(PageNumberPagination):
    page_size = 20
    max_page_size = 10000
